import abc

from task_processing.interfaces.task_executor import DefaultTaskConfigInterface


class Runner(metaclass=abc.ABCMeta):
    TASK_CONFIG_INTERFACE = DefaultTaskConfigInterface
    """
    The interface, specified as a PRecord of
    objects that you will be passing as task_configs to run
    """

    @abc.abstractmethod
    def run(self, task_config):
        pass

    @abc.abstractmethod
    def reconcile(self, task_config):
        pass

    @abc.abstractmethod
    def kill(self, task_id):
        pass



import json
import uuid
from typing import Any

from pyrsistent import field
from pyrsistent import freeze
from pyrsistent import m
from pyrsistent import PMap
from pyrsistent import pmap
from pyrsistent import PRecord

EVENT_KINDS = {'task', 'control'}
EVENT_TASK_ATTRS = {'task_id', 'task_config'}


class Event(PRecord):
    def __invariant__(self):
        if self.kind != 'task':
            return True, 'Not a task event'

        missing_attrs = EVENT_TASK_ATTRS - set(self.keys())
        if missing_attrs:
            return False, 'Missing task attributes: {}'.format(missing_attrs)

        return True, 'Task is valid'

    kind = field(type=str,
                 mandatory=True,
                 invariant=lambda x: (x in EVENT_KINDS,
                                      'kind not in {}'.format(EVENT_KINDS)))
    # we store timestamps as seconds since epoch.
    # use time.time() to generate
    timestamp = field(type=float, initial=0.0)
    # reference to platform-specific event object
    raw: Any = field(mandatory=True, initial=None)
    # free-form dictionary for stack-specific data
    extensions = field(type=PMap, initial=m(), factory=pmap)
    # is this the last event for a task?
    terminal = field(type=bool, initial=False)

    # task-specific fields
    # task_id this event pertains to
    task_id = field(type=(str, type(None)), initial=None)
    # task config dict that sourced the task this event refers to
    task_config = field(
        invariant=lambda x: (isinstance(x, PMap),
                             'task_config must inherit from PMap'),
        factory=lambda x: pmap(x) if not isinstance(x, PMap) else x,
        initial=m(),
    )
    # the task finished with exit code 0
    success = field(type=(bool, type(None)), initial=None)
    # platform-specific event type
    platform_type = field(type=(str, type(None)), initial=None)

    # control events
    message = field(type=(str, type(None)), initial=None)


def task_event(**kwargs):
    kwargs.setdefault('kind', 'task')
    return Event(**kwargs)


def control_event(**kwargs):
    kwargs.setdefault('kind', 'control')
    return Event(**kwargs)


def json_serializer(o):
    if isinstance(o, uuid.UUID):
        return o.hex
    return json.JSONEncoder.default(o)


def json_deserializer(dct):
    for k, v in dct.items():
        if k == "uuid":
            try:
                dct[k] = uuid.UUID(hex=v)
            except ValueError:
                dct[k] = freeze(v)
        else:
            dct[k] = freeze(v)
    return dct



import logging
import threading
import traceback
from queue import Queue

from task_processing.interfaces.task_executor import TaskExecutor

log = logging.getLogger(__name__)


class StatefulTaskExecutor(TaskExecutor):
    """
    """

    def __init__(self, downstream_executor, persister):
        self.downstream_executor = downstream_executor
        self.writer_queue = Queue()
        self.queue_for_processed_events = Queue()
        self.persister = persister
        worker_thread = threading.Thread(
            target=self.subscribe_to_updates_for_task
        )
        worker_thread.daemon = True
        worker_thread.start()

    def run(self, task_config):
        self.downstream_executor.run(task_config)

    def reconcile(self, task_config):
        self.downstream_executor.reconcile(task_config)

    def kill(self, task_id):
        return self.downstream_executor.kill(task_id)

    def status(self, task_id):
        return sorted(
            self.persister.read(task_id),
            key=lambda x: x['timestamp']
        )

    def stop(self):
        return self.downstream_executor.stop()

    def get_event_queue(self):
        return self.queue_for_processed_events

    def subscribe_to_updates_for_task(self):
        while True:
            result = self.downstream_executor.get_event_queue().get()
            try:
                self.persister.write(event=result)
            except Exception:
                log.error(traceback.format_exc())
            self.queue_for_processed_events.put(result)
            self.downstream_executor.get_event_queue().task_done()



from task_processing.interfaces.runner import Runner
from task_processing.runners.sync import Sync


class Promise(Runner):
    def __init__(self, executor, futures_executor):
        self.TASK_CONFIG_INTERFACE = executor.TASK_CONFIG_INTERFACE
        self._futures_executor = futures_executor
        self._runner = Sync(executor)

    def run(self, task_config):
        """Schedules execution of the supplied task

        :param task_config: An object satisfying the TASK_CONFIG_INTERFACE

        :return: A Future object representing the execution of the task.
        """
        return self._futures_executor.submit(self._runner.run, task_config)

    def kill(self, task_id):
        return self._runner.kill(task_id)

    def reconcile(self, task_config):
        self._runner.reconcile(task_config)

    def stop(self):
        self._runner.stop()



import logging
import queue
import threading
import time
from queue import Queue
from typing import Collection
from typing import Optional

from kubernetes import watch
from kubernetes.client import V1Container
from kubernetes.client import V1ObjectMeta
from kubernetes.client import V1Pod
from kubernetes.client import V1PodSpec
from kubernetes.client import V1ResourceRequirements
from kubernetes.client.exceptions import ApiException
from pyrsistent import pmap
from pyrsistent import v
from pyrsistent.typing import PMap

from task_processing.interfaces import TaskExecutor
from task_processing.interfaces.event import Event
from task_processing.interfaces.event import task_event
from task_processing.plugins.kubernetes.kube_client import KubeClient
from task_processing.plugins.kubernetes.task_config import KubernetesTaskConfig
from task_processing.plugins.kubernetes.types import KubernetesTaskMetadata
from task_processing.plugins.kubernetes.types import KubernetesTaskState
from task_processing.plugins.kubernetes.types import PodEvent
from task_processing.plugins.kubernetes.utils import get_kubernetes_env_vars
from task_processing.plugins.kubernetes.utils import get_kubernetes_volume_mounts
from task_processing.plugins.kubernetes.utils import get_pod_volumes
from task_processing.plugins.kubernetes.utils import get_security_context_for_capabilities

logger = logging.getLogger(__name__)

POD_WATCH_THREAD_JOIN_TIMEOUT_S = 1.0
POD_EVENT_THREAD_JOIN_TIMEOUT_S = 1.0
QUEUE_GET_TIMEOUT_S = 0.5
SUPPORTED_POD_MODIFIED_EVENT_PHASES = {
    "Failed",
    "Running",
    "Succeeded",
    "Unknown",
}


class KubernetesPodExecutor(TaskExecutor):
    TASK_CONFIG_INTERFACE = KubernetesTaskConfig

    def __init__(
        self,
        namespace: str,
        kubeconfig_path: Optional[str] = None,
        task_configs: Optional[Collection[KubernetesTaskConfig]] = [],
    ) -> None:
        self.kube_client = KubeClient(kubeconfig_path=kubeconfig_path)
        self.namespace = namespace
        self.stopping = False
        self.task_metadata: PMap[str, KubernetesTaskMetadata] = pmap()

        self.task_metadata_lock = threading.RLock()
        if task_configs:
            for task_config in task_configs:
                self._initialize_existing_task(task_config)

        # we have two queues since we need to process Pod Events into something that we can place
        # onto the queue that any application (e.g., something like Tron or Jolt) can actually use
        # and we've opted to not do that processing in the Pod event watcher thread so as to keep
        # that logic for the threads that operate on them as simple as possible and to make it
        # possible to cleanly shutdown both of these.
        self.pending_events: "Queue[PodEvent]" = Queue()
        self.event_queue: "Queue[Event]" = Queue()

        # TODO(TASKPROC-243): keep track of resourceVersion so that we can continue event processing
        # from where we left off on restarts
        self.watch = watch.Watch()
        self.pod_event_watch_thread = threading.Thread(
            target=self._pod_event_watch_loop,
            # ideally this wouldn't be a daemon thread, but a watch.Watch() only checks
            # if it should stop after receiving an event - and it's possible that we
            # have periods with no events so instead we'll attempt to stop the watch
            # and then join() with a small timeout to make sure that, if we shutdown
            # with the thread alive, we did not drop any events
            daemon=True,
        )
        self.pod_event_watch_thread.start()

        self.pending_event_processing_thread = threading.Thread(
            target=self._pending_event_processing_loop,
        )
        self.pending_event_processing_thread.start()

    def _initialize_existing_task(self, task_config: KubernetesTaskConfig) -> None:
        """ Generates task_metadata in UNKNOWN state for an existing KubernetesTaskConfig.
            Used during initialization or recovery for a task"""
        pod_name = task_config.pod_name
        logger.debug(
            f"Initializing task metadata for known pod {pod_name}"
        )
        # We are initiating with UNKNOWN state on initial load, then leave
        # matching tasks to running/completed pods in reconcile()
        with self.task_metadata_lock:
            self.task_metadata = self.task_metadata.set(
                pod_name,
                KubernetesTaskMetadata(
                    task_config=task_config,
                    node_name='UNKNOWN',
                    task_state=KubernetesTaskState.TASK_UNKNOWN,
                    task_state_history=v(
                        (KubernetesTaskState.TASK_UNKNOWN, time.time())
                    ),
                )
            )

    def _pod_event_watch_loop(self) -> None:
        logger.debug(f"Starting watching Pod events for namespace={self.namespace}.")
        # TODO(TASKPROC-243): we'll need to correctly handle resourceVersion expiration for the case
        # where the gap between task_proc shutting down and coming back up is long enough for data
        # to have expired from etcd as well as actually restarting from the last resourceVersion in
        # case of an exception
        # TODO: Do LIST + WATCH if we're not starting from a known/good resourceVersion to
        # guarantee that we always get ordered events (starting from a resourceVersion of 0)
        # has no such guarantees that the initial events will be ordered in any meaningful way
        # see: https://github.com/kubernetes/kubernetes/issues/74022
        while not self.stopping:
            try:
                for pod_event in self.watch.stream(
                    self.kube_client.core.list_namespaced_pod,
                    self.namespace
                ):
                    # it's possible that we've received an event after we've already set the stop
                    # flag since Watch streams block forever, so re-check if we've stopped before
                    # queueing any pending events
                    if not self.stopping:
                        logger.debug("Adding Pod event to pending event queue.")
                        self.pending_events.put(pod_event)
                    else:
                        break
            except ApiException as e:
                if not self.stopping:
                    if not self.kube_client.maybe_reload_on_exception(exception=e):
                        logger.exception(
                            "Unhandled API exception while watching Pod events - restarting watch!"
                        )
            except Exception:
                # we want to avoid a potentially misleading log message should we encounter
                # an exception when we want to shutdown this thread since nothing of value
                # will be lost anyway (example scenario: we've called stop() on an idle cluster
                # but the Watch is still blocking and eventually gets disconnected)
                if not self.stopping:
                    logger.exception(
                        "Exception encountered while watching Pod events - restarting watch!")
        logger.debug("Exiting Pod event watcher - stop requested.")

    def __handle_deleted_pod_event(self, event: PodEvent) -> None:
        pod = event["object"]
        pod_name = pod.metadata.name
        task_metadata = self.task_metadata[pod_name]
        raw_event = event['raw_object']

        logger.info(f"Removing {pod_name} from state and emitting 'killed' event.")

        self.task_metadata = self.task_metadata.discard(pod_name)
        self.event_queue.put(
            task_event(
                task_id=pod_name,
                terminal=True,
                success=False,
                timestamp=time.time(),
                raw=raw_event,
                task_config=task_metadata.task_config,
                platform_type="killed"
            )
        )

    def __handle_modified_pod_event(self, event: PodEvent) -> None:
        pod = event["object"]
        self.__update_modified_pod(pod=pod, event=event)

    def __update_modified_pod(self, pod: V1Pod, event: Optional[PodEvent]) -> None:
        """ Called during reconciliation and normal event handling """
        pod_name = pod.metadata.name
        task_metadata = self.task_metadata[pod_name]

        raw_event = event['raw_object'] if event else None

        if pod.status.phase not in SUPPORTED_POD_MODIFIED_EVENT_PHASES:
            logger.debug(
                f"Got a MODIFIED event for {pod_name} for unhandled phase: "
                f"{pod.status.phase} - ignoring."
            )
            return

        if (
            pod.status.phase == "Succeeded"
            and task_metadata.task_state is not KubernetesTaskState.TASK_FINISHED
        ):
            logger.info(
                f"Removing {pod_name} from state and emitting 'finished' event.",
            )
            self.task_metadata = self.task_metadata.discard(pod_name)
            self.event_queue.put(
                task_event(
                    task_id=pod_name,
                    terminal=True,
                    success=True,
                    timestamp=time.time(),
                    raw=raw_event,
                    task_config=task_metadata.task_config,
                    platform_type="finished"
                )
            )
            return

        elif (
            pod.status.phase == "Failed"
            and task_metadata.task_state is not KubernetesTaskState.TASK_FAILED
        ):
            logger.info(f"Removing {pod_name} from state and emitting 'failed' event.")
            self.task_metadata = self.task_metadata.discard(pod_name)
            self.event_queue.put(
                task_event(
                    task_id=pod_name,
                    terminal=True,
                    success=False,
                    timestamp=time.time(),
                    raw=raw_event,
                    task_config=task_metadata.task_config,
                    platform_type="failed"
                )
            )
            return

        elif (
            pod.status.phase == "Running"
            and task_metadata.task_state is not KubernetesTaskState.TASK_RUNNING
        ):
            logger.info(f"Successfully launched {pod_name}, emitting 'running' event.")
            self.task_metadata = self.task_metadata.set(
                pod_name,
                task_metadata.set(
                    node_name=pod.spec.node_name,
                    task_state=KubernetesTaskState.TASK_RUNNING,
                    task_state_history=task_metadata.task_state_history.append(
                        (KubernetesTaskState.TASK_RUNNING, time.time()),
                    )
                )
            )
            self.event_queue.put(
                task_event(
                    task_id=pod_name,
                    terminal=False,
                    timestamp=time.time(),
                    raw=raw_event,
                    task_config=task_metadata.task_config,
                    platform_type="running"
                )
            )
            return

        # XXX: figure out how to handle this correctly (and when this actually
        # happens - we were unable to cajole k8s into giving us an event with an Unknown
        # phase)
        elif (
            pod.status.phase == "Unknown"
            and task_metadata.task_state is not KubernetesTaskState.TASK_LOST
        ):
            logger.info(
                f"Got a MODIFIED event for {pod_name} with unknown phase, host likely "
                "unexpectedly died"
            )
            self.task_metadata = self.task_metadata.set(
                pod_name,
                task_metadata.set(
                    node_name=pod.spec.node_name,
                    task_state=KubernetesTaskState.TASK_LOST,
                    task_state_history=task_metadata.task_state_history.append(
                        (KubernetesTaskState.TASK_LOST, time.time()),
                    )
                )
            )
            self.event_queue.put(
                task_event(
                    task_id=pod_name,
                    terminal=False,
                    timestamp=time.time(),
                    raw=raw_event,
                    task_config=task_metadata.task_config,
                    platform_type="lost"
                )
            )
            return

        logger.info(
            f"Ignoring MODIFIED event for {pod_name} as it did not result "
            "in a state transition",
        )

    def _process_pod_event(self, event: PodEvent) -> None:
        """
        Router for handling Pod events based on event type. Currently, we only look at
        MODIFIED and DELETED events (we ignore ADDED since we handle all processing for
        those when we create Pods in the first place).

        NOTE: if we get multiple pod events while a Pod is running, we ignore all
        but the first one to cause us to transition to a new state to make processing
        easier since our watch might result in events with no state transitions and these
        don't have any value for us.

        """
        pod = event["object"]
        pod_name = pod.metadata.name
        with self.task_metadata_lock:
            if pod_name not in self.task_metadata:
                logger.info(f"Ignoring event for {pod_name} - Pod not tracked by task_processing.")
                return None

            # this should only ever be run for Pods that were killed either by operator
            # action (e.g., kubectl delete pod) or by calling kill() - completed/failed
            # Pods will be removed from task_processing's metadata store
            elif event["type"] == "DELETED":
                self.__handle_deleted_pod_event(event)

            elif event["type"] == "MODIFIED":
                self.__handle_modified_pod_event(event)

            else:
                logger.warning(f"Got unknown event type for {pod_name}: {event['type']}")

    def _pending_event_processing_loop(self) -> None:
        """
        Run in a thread to process PodEvents enqueued by the k8s event watcher thread.
        """
        logger.debug("Starting Pod event processing.")
        event = None
        while not self.stopping or not self.pending_events.empty():
            try:
                event = self.pending_events.get(timeout=QUEUE_GET_TIMEOUT_S)
                self._process_pod_event(event)
            except queue.Empty:
                logger.debug(
                    f"Pending event queue remained empty after {QUEUE_GET_TIMEOUT_S} seconds.",
                )
                # we explicitly continue here as this is the only code path that doesn't
                # need to call task_done()
                continue
            except Exception:
                logger.exception(
                    f"Exception encountered proccessing Pod events - skipping event {event} and "
                    "restarting processing!",
                )

            try:
                self.pending_events.task_done()
            except ValueError:
                # this should never happen since the only codepath that should ever get
                # here is after a successful get() - but just in case, we don't want to
                # have this thread die because of this
                logger.error("task_done() called on pending events queue too many times!")

        logger.debug("Exiting Pod event processing - stop requested.")

    def run(self, task_config: KubernetesTaskConfig) -> Optional[str]:
        try:
            container = V1Container(
                image=task_config.image,
                # XXX: we were initially planning on using the name from KubernetesTaskConfig here,
                # but its too easy to go over the length limit for container names (63 characters),
                # so we're just hardcoding something for now since container names aren't used for
                # anything at the moment
                name="main",
                command=["/bin/sh", "-c"],
                args=[task_config.command],
                security_context=get_security_context_for_capabilities(
                    cap_add=task_config.cap_add,
                    cap_drop=task_config.cap_drop,
                ),
                resources=V1ResourceRequirements(
                    limits={
                        "cpu": task_config.cpus,
                        "memory": f"{task_config.memory}Mi",
                        "ephemeral-storage": f"{task_config.disk}Mi",
                    }
                ),
                env=get_kubernetes_env_vars(
                    task_config.environment,
                    task_config.secret_environment,
                ),
                volume_mounts=get_kubernetes_volume_mounts(task_config.volumes)
            )
            pod = V1Pod(
                metadata=V1ObjectMeta(
                    name=task_config.pod_name,
                    namespace=self.namespace
                ),
                spec=V1PodSpec(
                    restart_policy=task_config.restart_policy,
                    containers=[container],
                    volumes=get_pod_volumes(task_config.volumes)
                ),
            )
        except Exception:
            logger.exception(f"Unable to create PodSpec for {task_config.pod_name}")
            return None

        # we need to lock here since there will be other threads updating this metadata in response
        # to k8s events
        with self.task_metadata_lock:
            self.task_metadata = self.task_metadata.set(
                task_config.pod_name,
                KubernetesTaskMetadata(
                    task_config=task_config,
                    task_state=KubernetesTaskState.TASK_PENDING,
                    task_state_history=v(
                        (KubernetesTaskState.TASK_PENDING, time.time())
                    ),
                ),
            )

        if self.kube_client.create_pod(
            namespace=self.namespace,
            pod=pod,
        ):
            return task_config.pod_name

        with self.task_metadata_lock:
            # if we weren't able to create this pod, then we shouldn't keep it around in
            # task_metadata
            self.task_metadata = self.task_metadata.discard(task_config.pod_name)

        return None

    def reconcile(self, task_config: KubernetesTaskConfig) -> None:
        pod_name = task_config.pod_name
        try:
            pod = self.kube_client.get_pod(namespace=self.namespace, pod_name=pod_name)
        except Exception:
            logger.exception(f"Hit an exception attempting to fetch pod {pod_name}")
            pod = None

        if pod_name not in self.task_metadata:
            self._initialize_existing_task(task_config)

        with self.task_metadata_lock:
            task_metadata = self.task_metadata[pod_name]
            self.task_metadata = self.task_metadata.set(
                pod_name,
                task_metadata.set(
                    task_config=task_config
                )
            )

            if not pod:
                # Pod has gone away while restarting
                logger.info(
                    f"Pod {pod_name} for task {task_config.name} was no longer found. "
                    "Marking as LOST"
                )
                self.task_metadata = self.task_metadata.set(
                    pod_name,
                    task_metadata.set(
                        task_state=KubernetesTaskState.TASK_LOST,
                        task_state_history=task_metadata.task_state_history.append(
                            (KubernetesTaskState.TASK_LOST, time.time()),
                        )
                    )
                )
                self.event_queue.put(
                    task_event(
                        task_id=pod_name,
                        terminal=False,
                        timestamp=time.time(),
                        raw=None,
                        task_config=task_metadata.task_config,
                        platform_type="lost"
                    )
                )
            else:
                # Treat like a modified pod
                self.__update_modified_pod(pod=pod, event=None)

    def kill(self, task_id: str) -> bool:
        """
        Terminate a Pod by name.

        This function will request that Kubernetes delete the named Pod and will return
        True if the Pod termination request was succesfully emitted or False otherwise.
        """
        # NOTE: we're purposely not removing this task from `task_metadata` as we want
        # to handle that with the Watch that we'll set to monitor each Pod for events.
        # TODO(TASKPROC-242): actually handle termination events
        return self.kube_client.terminate_pod(
            namespace=self.namespace,
            pod_name=task_id,
        )

    def stop(self) -> None:
        logger.debug("Preparing to stop all KubernetesPodExecutor threads.")
        self.stopping = True

        logger.debug("Signaling Pod event Watch to stop streaming events...")
        # make sure that we've stopped watching for events before calling join() - otherwise,
        # join() will block until we hit the configured timeout (or forever with no timeout).
        self.watch.stop()
        # timeout arbitrarily chosen - we mostly just want to make sure that we have a small
        # grace period to flush the current event to the pending_events queue as well as
        # any other clean-up  - it's possible that after this join() the thread is still alive
        # but in that case we can be reasonably sure that we're not dropping any data.
        self.pod_event_watch_thread.join(timeout=POD_WATCH_THREAD_JOIN_TIMEOUT_S)

        logger.debug("Waiting for all pending PodEvents to be processed...")
        # once we've stopped updating the pending events queue, we then wait until we're done
        # processing any events we've received - this will wait until task_done() has been
        # called for every item placed in this queue
        self.pending_events.join()
        logger.debug("All pending PodEvents have been processed.")
        # and then give ourselves time to do any post-stop cleanup
        self.pending_event_processing_thread.join(timeout=POD_EVENT_THREAD_JOIN_TIMEOUT_S)

        logger.debug("Done stopping KubernetesPodExecutor!")

    def get_event_queue(self) -> "Queue[Event]":
        return self.event_queue



from queue import Empty
from queue import Full
from threading import Thread

from task_processing.interfaces.runner import Runner


class Subscription(Runner):
    def __init__(self, executor, queue):
        self.executor = executor
        self.TASK_CONFIG_INTERFACE = executor.TASK_CONFIG_INTERFACE
        self.event_queue = queue
        self.stopping = False
        self.producer_t = Thread(target=self.event_producer)
        self.producer_t.daemon = True
        self.producer_t.start()

    def event_producer(self):
        executor_queue = self.executor.get_event_queue()
        while True:
            if self.stopping:
                return
            try:
                event = executor_queue.get(block=True, timeout=1)
                self.event_queue.put(event, False)
            except Empty:
                pass
            except Full:
                pass

    def run(self, task_config):
        return self.executor.run(task_config)

    def kill(self, task_id):
        return self.executor.kill(task_id)

    def reconcile(self, task_config):
        self.executor.reconcile(task_config)

    def stop(self):
        self.executor.stop()
        self.stopping = True
        self.producer_t.join()



try:
    import yelp_meteorite
    METRICS_ENABLED = True
except Exception:
    METRICS_ENABLED = False


class _DummyMetricType:
    """
    Emulates a yelp_meteorite counter, gauge, or timer
    """
    def count(*args, **kwargs):
        pass

    set = count
    record = count
    start = count
    stop = count


_dummy_metric = _DummyMetricType()
_registered_metrics: dict = {}


def create_counter(name, dimensions={}):
    if not METRICS_ENABLED:
        return

    if name not in _registered_metrics:
        counter = yelp_meteorite.create_counter(
            name, default_dimensions=dimensions)
        _registered_metrics[name] = counter


def create_timer(name, dimensions={}):
    if not METRICS_ENABLED:
        return

    if name not in _registered_metrics:
        timer = yelp_meteorite.create_timer(
            name, default_dimensions=dimensions)
        _registered_metrics[name] = timer


def get_metric(name):
    if METRICS_ENABLED:
        return _registered_metrics.get(name)
    else:
        return _dummy_metric



from .kubernetes_pod_executor import KubernetesPodExecutor


TASK_PROCESSING_PLUGIN = 'kubernetes_plugin'


def register_plugin(registry):
    return registry \
        .register_task_executor('kubernetes', KubernetesPodExecutor)



import logging
import os
from http import HTTPStatus
from typing import Optional

from kubernetes import client as kube_client
from kubernetes import config as kube_config
from kubernetes.client.exceptions import ApiException
from kubernetes.client.models.v1_pod import V1Pod

logger = logging.getLogger(__name__)

DEFAULT_ATTEMPTS = 2


class ExceededMaxAttempts(Exception):
    pass


class KubeClient:
    def __init__(self, kubeconfig_path: Optional[str] = None) -> None:
        kubeconfig_path = kubeconfig_path or os.environ.get("KUBECONFIG")
        if kubeconfig_path is None:
            raise ValueError(
                "No kubeconfig specified: set a KUBECONFIG environment variable "
                "or pass a value for `kubeconfig_path`!"
            )

        self.kubeconfig_path = kubeconfig_path
        self.kubecontext = os.environ.get("KUBECONTEXT")

        kube_config.load_kube_config(
            config_file=self.kubeconfig_path,
            context=self.kubecontext
        )

        # any Kubernetes APIs that we use should be added as members here (much like as we
        # do in the KubeClient class in PaaSTA) to ensure that they're only used after we've
        # loaded the relevant kubeconfig. Additionally, they should also be recreated in
        # reload_kubeconfig() so that changes in certs are picked up.
        self.core = kube_client.CoreV1Api()

    def reload_kubeconfig(self) -> None:
        """
        Helper meant to be called after a 401 from apiserver.

        Since we rotate certs regularly internally and the kubernetes clientlib doesn't
        automatically watch for changes to the certs specified by the kubeconfig file, we
        need to manually reload said file as well as recreate any API classes (as otherwise
        they don't pick up the new configuration)
        """
        kube_config.load_kube_config(
            config_file=self.kubeconfig_path,
            context=self.kubecontext
        )
        self.core = kube_client.CoreV1Api()

    def maybe_reload_on_exception(self, exception: Exception) -> bool:
        """
        Small wrapper around KubeClient::reload_kubeconfig() to centralize when to reload kubeconfig
        on exceptions.

        Currently, the only exception that will cause a reload is an ApiException with an
        HTTP 401 (unauthorized) status.
        """
        if isinstance(exception, ApiException):
            if exception.status == HTTPStatus.UNAUTHORIZED.value:
                logger.info(
                    "Recieved UNAUTHORIZED response from apiserver - assuming certs have "
                    "expired and reloading."
                )
                self.reload_kubeconfig()
                return True
            logger.info(f"Recieved HTTP {exception.status} from apiserver, not reloading certs.")
        return False

    def terminate_pod(
        self,
        namespace: str,
        pod_name: str,
        attempts: int = DEFAULT_ATTEMPTS,
    ) -> bool:
        """
        Wrapper around delete_namespaced_pod() in the kubernetes clientlib that adds in
        retrying on ApiExceptions.

        Returns True on success, False otherwise.
        """
        while attempts:
            try:
                logger.info(f"Attempting to terminate {pod_name}")
                self.core.delete_namespaced_pod(
                    name=pod_name,
                    namespace=namespace,
                    # attempt to delete immediately - Pods launched by task_processing
                    # shouldn't need time to clean-up/drain
                    grace_period_seconds=0,
                    # this is the default, but explcitly request background deletion of releated
                    # objects. see:
                    # https://kubernetes.io/docs/concepts/workloads/controllers/garbage-collection/
                    propagation_policy="Background"
                )
                # this is not ideal, but the k8s clientlib should return the status of the request
                # as a string that is either "Success" or "Failure" (or we could potentially use
                # `code` instead but it's not exactly documented what HTTP return codes will be
                # used)...however, due to https://github.com/kubernetes-client/python/issues/1523
                # the V1Status.status returned by delete_namespaced_pod is not accurate and thus
                # we assume that if no exception was thrown, we were able to successfully send
                # the termination request.
                return True
            except ApiException as e:
                if not self.maybe_reload_on_exception(exception=e) and attempts:
                    logger.exception(
                        f"Failed to request termination for {pod_name} due to unhandled API "
                        "exception, retrying."
                    )
                attempts -= 1
            except Exception:
                logger.exception(
                    f"Failed to request termination for {pod_name} due to unhandled exception."
                )
                return False

        logger.info(f"Ran out of retries attempting to request termination of {pod_name}.")
        return False

    def create_pod(
        self,
        namespace: str,
        pod: V1Pod,
        attempts: int = DEFAULT_ATTEMPTS,
    ) -> bool:
        """
        Wrapper around create_namespaced_pod() in the kubernetes clientlib that adds in
        retrying on ApiExceptions.

        Returns True on success, False otherwise.
        """
        while attempts:
            try:
                self.core.create_namespaced_pod(
                    namespace=namespace,
                    body=pod,
                )
                logger.info(f"Successfully created pod {pod.metadata.name}")
                return True
            except ApiException as e:
                if not self.maybe_reload_on_exception(exception=e) and attempts:
                    logger.exception(
                        f"Failed to create {pod.metadata.name} due to unhandled API exception, "
                        "retrying."
                    )
                attempts -= 1
            except Exception:
                logger.exception(
                    f"Failed to create {pod.metadata.name} due to unhandled "
                    "exception."
                )
                return False

        logger.info(f"Ran out of retries attempting to create {pod.metadata.name}.")
        return False

    def get_pod(
        self, namespace: str, pod_name: str, attempts: int = DEFAULT_ATTEMPTS,
    ) -> Optional[V1Pod]:
        max_attempts = attempts
        while attempts:
            try:
                pod = self.core.read_namespaced_pod(
                    namespace=namespace, name=pod_name,
                )
                return pod
            except ApiException as e:
                # Unknown pod throws ApiException w/ 404
                if e.status == 404:
                    logger.info(f"Found no pods matching {pod_name}.")
                    return None
                if not self.maybe_reload_on_exception(exception=e) and attempts:
                    logger.debug(
                        f"Failed to fetch pod {pod_name} due to unhandled API exception, retrying",
                        exc_info=True
                    )
                attempts -= 1
            except Exception:
                logger.exception(
                    f"Failed to fetch pod {pod_name} due to unhandled exception."
                )
                raise
        logger.info(f"Ran out of retries attempting to fetch pod {pod_name}.")
        raise ExceededMaxAttempts(f'Retried fetching pod {pod_name} {max_attempts} times.')



import logging
import sys
import time
from queue import Queue
from threading import Lock
from threading import Thread
from urllib.parse import urlparse

import requests
from pyrsistent import field
from pyrsistent import m
from pyrsistent import PMap
from pyrsistent import pmap
from pyrsistent import PRecord
from pyrsistent import v

from task_processing.interfaces.task_executor import TaskExecutor


log = logging.getLogger(__name__)
logging.getLogger("urllib3").setLevel(logging.WARNING)

# Read task log in 4K chunks
TASK_LOG_CHUNK_LEN = 4096
DEFAULT_FORMAT = '{task_id}[{container_id}@{agent}]: {line}'
LOG_REQUEST_TIMEOUT = 5  # seconds


class LogMetadata(PRecord):
    log_url = field(type=str, initial='')
    log_path = field(type=str, initial='')
    log_offsets = field(type=PMap,
                        factory=pmap,
                        initial=pmap({'stdout': 0, 'stderr': 0}))
    container_id = field(type=str, initial='')
    executor_id = field(type=str, initial='')


def standard_handler(task_id, message, stream):
    print(message, file=sys.stderr if stream == 'stderr' else sys.stdout)


class MesosLoggingExecutor(TaskExecutor):
    def __init__(
        self,
        downstream_executor,
        handler=standard_handler,
        format_string=DEFAULT_FORMAT,
    ):
        self.downstream_executor = downstream_executor
        self.TASK_CONFIG_INTERFACE = downstream_executor.TASK_CONFIG_INTERFACE
        self.handler = handler
        self.format_string = format_string

        self.src_queue = downstream_executor.get_event_queue()
        self.dest_queue = Queue()
        self.stopping = False

        self.staging_tasks = m()
        self.running_tasks = m()
        self.done_tasks = v()

        # A lock is needed to synchronize logging and event processing
        self.task_lock = Lock()

        self.event_thread = Thread(target=self.event_loop)
        self.event_thread.daemon = True
        self.event_thread.start()

        self.logging_thread = Thread(target=self.logging_loop)
        self.logging_thread.daemon = True
        self.logging_thread.start()

    def log_line(self, stream, line, task_id, container_id, agent):
        formatted_line = self.format_string.format(
            task_id=task_id,
            container_id=container_id,
            agent=agent,
            line=line,
        )
        self.handler(task_id, formatted_line, stream)

    def set_task_log_path(self, task_id):
        log_md = self.running_tasks[task_id]
        if log_md.log_url is None:
            log.error(f"No log_url available for {task_id}")
            return
        try:
            response = requests.get(
                log_md.log_url + '/files/debug',
                timeout=LOG_REQUEST_TIMEOUT,
            ).json()
        except Exception as e:
            log.error("Failed to fetch files {error}".format(error=e))
            return

        for key in response.keys():
            if log_md.executor_id in key and log_md.container_id in key:
                with self.task_lock:
                    self.running_tasks = self.running_tasks.set(
                        task_id,
                        log_md.set(log_path=key),
                    )
                break

    def stream_task_log(self, task_id):
        if self.running_tasks[task_id].log_path == "":
            self.set_task_log_path(task_id)

        # Abort in case the log path discovery was not successful
        log_md = self.running_tasks[task_id]
        if log_md.log_path == "":
            return

        offsets = {
            'stdout': log_md.log_offsets['stdout'],
            'stderr': log_md.log_offsets['stderr']
        }
        agent = urlparse(log_md.log_url).hostname

        for f in ['stdout', 'stderr']:
            offset = offsets[f]
            log_path = log_md.log_path + "/" + f
            while True:
                payload = {
                    'path': log_path,
                    'length': str(TASK_LOG_CHUNK_LEN),
                    'offset': str(offset)
                }

                try:
                    response = requests.get(
                        log_md.log_url + '/files/read',
                        params=payload,
                        timeout=LOG_REQUEST_TIMEOUT,
                    ).json()

                    log_length = len(response['data'])
                    for line in response['data'].splitlines():
                        self.log_line(
                            stream=f,
                            line=line,
                            task_id=task_id,
                            container_id=log_md.container_id,
                            agent=agent,
                        )
                except Exception as e:
                    log.error("Failed to get {path}@{agent} {error}".format(
                        path=log_path, agent=agent, error=e
                    ))
                    break

                offset = offset + log_length
                # Stop if there is no more data
                if log_length < TASK_LOG_CHUNK_LEN:
                    break
            # Update offset of this stream
            offsets[f] = offset

        # Update both offsets for the task
        with self.task_lock:
            self.running_tasks = self.running_tasks.set(
                task_id,
                log_md.set(log_offsets=pmap(offsets)),
            )

    # process downstream events
    def event_loop(self):
        while True:
            while not self.src_queue.empty():
                e = self.src_queue.get()
                self.dest_queue.put(e)
                self.src_queue.task_done()

                # Record the base log url
                if e.kind == 'task' and e.platform_type == 'staging':
                    if e.task_id in self.staging_tasks:
                        continue
                    url = extract_url_from_offer(e.raw.offer)
                    self.staging_tasks = self.staging_tasks.set(e.task_id, url)

                if e.kind == 'task' and e.platform_type == 'running':
                    if e.task_id not in self.staging_tasks:
                        log.info(
                            f"Task {e.task_id} already running, not fetching logs")
                        continue

                    url = self.staging_tasks[e.task_id]
                    self.staging_tasks = self.staging_tasks.discard(e.task_id)

                    # Simply pass the needed fields and let the logging thread
                    # to take care of the slow path discovery.
                    container_id = e.raw.container_status.container_id.value
                    executor_id = e.raw.executor_id.value
                    with self.task_lock:
                        self.running_tasks = self.running_tasks.set(
                            e.task_id,
                            LogMetadata(
                                log_url=url,
                                container_id=container_id,
                                executor_id=executor_id
                            )
                        )

                # Fetch the last log and remove the entry if the task is active
                if e.kind == 'task' and e.terminal:
                    with self.task_lock:
                        if e.task_id in self.running_tasks:
                            self.done_tasks = self.done_tasks.append(e.task_id)

            if self.stopping:
                return

            time.sleep(1)

    def logging_loop(self):
        while True:
            # grab logs
            for task_id in self.running_tasks.keys():
                self.stream_task_log(task_id)

            while len(self.done_tasks):
                task_id = self.done_tasks[0]
                self.stream_task_log(task_id)
                with self.task_lock:
                    self.done_tasks = self.done_tasks.remove(task_id)
                    self.running_tasks = self.running_tasks.discard(task_id)

            if self.stopping:
                return

            time.sleep(10)

    def run(self, task_config):
        self.downstream_executor.run(task_config)

    def stop(self):
        self.downstream_executor.stop()
        self.stopping = True
        self.event_thread.join()
        self.logging_thread.join()

    def get_event_queue(self):
        return self.dest_queue

    def reconcile(self, task_config):
        self.downstream_executor.reconcile(task_config)

    def kill(self, task_id):
        return self.downstream_executor.kill(task_id)


def extract_url_from_offer(offer):
    try:
        url = offer.url.scheme + '://' + \
            offer.url.address.ip + ':' + \
            str(offer.url.address.port)
    except Exception as exc:
        log.error(
            f"Error decoding the url for this offer: {offer.url}. "
            f"Setting to None. Exception: {exc}"
        )
        url = None
    return url



from enum import auto
from enum import unique
from typing import Any
from typing import Dict
from typing import Tuple

from kubernetes.client import V1Pod
from pyrsistent import field
from pyrsistent import PRecord
from pyrsistent import PVector
from pyrsistent import pvector
from pyrsistent.typing import PVector as PVectorType
from typing_extensions import TypedDict

from task_processing.plugins.kubernetes.task_config import KubernetesTaskConfig
from task_processing.utils import AutoEnum


class DockerVolume(TypedDict):
    host_path: str
    container_path: str
    mode: str  # XXX: Literal["RO", "RW"] once we drop older Python support


class SecretEnvSource(TypedDict):
    secret_name: str  # full name of k8s secret resource
    key: str


class PodEvent(TypedDict):
    # there are only 3 possible types for Pod events: ADDED, DELETED, MODIFIED
    # XXX: this should be typed as Literal["ADDED", "DELETED", "MODIFIED"] once we drop support
    # for older Python versions
    type: str
    object: V1Pod
    # this is just the dict-ified version of object - but it's too big to type here
    raw_object: Dict[str, Any]


@unique
class KubernetesTaskState(AutoEnum):
    TASK_PENDING = auto()
    TASK_RUNNING = auto()
    TASK_FINISHED = auto()
    TASK_FAILED = auto()
    TASK_UNKNOWN = auto()
    TASK_LOST = auto()


class KubernetesTaskMetadata(PRecord):
    # what box this task/Pod was scheduled onto
    node_name: str = field(type=str, initial='')

    # the config used to launch this task/Pod
    task_config: KubernetesTaskConfig = field(type=KubernetesTaskConfig, mandatory=True)

    task_state: KubernetesTaskState = field(type=KubernetesTaskState, mandatory=True)
    # List of state to when that state was entered (stored as a timestamp)
    task_state_history: PVectorType[Tuple[KubernetesTaskState, float]] = field(
        type=PVector, factory=pvector, mandatory=True)



import time
from typing import List

import addict
from pyrsistent import thaw

from task_processing.interfaces.event import Event
from task_processing.interfaces.event import task_event
from task_processing.plugins.mesos.task_config import MesosTaskConfig

# https://github.com/apache/mesos/blob/master/include/mesos/mesos.proto


def make_mesos_container_info(task_config: MesosTaskConfig) -> addict.Dict:
    container_info = addict.Dict(
        type=task_config.containerizer,
        volumes=thaw(task_config.volumes),
    )
    port_mappings = [addict.Dict(
        host_port=task_config.ports[0]['begin'], container_port=8888)]
    if container_info.type == 'DOCKER':
        container_info.docker = addict.Dict(
            image=task_config.image,
            network='BRIDGE',
            port_mappings=port_mappings,
            parameters=thaw(task_config.docker_parameters),
            force_pull_image=(not task_config.use_cached_image),
        )
    elif container_info.type == 'MESOS':
        container_info.network_infos = addict.Dict(port_mappings=port_mappings)
        # For this to work, image_providers needs to be set to 'docker' on mesos agents (as opposed
        # to 'appc' or 'oci'; we're still running docker images, we're just
        # using the UCR to do it).
        if 'image' in task_config:
            container_info.mesos.image = addict.Dict(
                type='DOCKER',  # not 'APPC' or 'OCI'
                docker=addict.Dict(name=task_config.image),
                cached=task_config.use_cached_image,
            )
    return container_info


def make_mesos_resources(
    task_config: MesosTaskConfig,
    role: str,
) -> List[addict.Dict]:
    return [
        addict.Dict(
            name='cpus',
            type='SCALAR',
            role=role,
            scalar=addict.Dict(value=task_config.cpus),
        ),
        addict.Dict(
            name='mem',
            type='SCALAR',
            role=role,
            scalar=addict.Dict(value=task_config.mem)
        ),
        addict.Dict(
            name='disk',
            type='SCALAR',
            role=role,
            scalar=addict.Dict(value=task_config.disk)
        ),
        addict.Dict(
            name='gpus',
            type='SCALAR',
            role=role,
            scalar=addict.Dict(value=task_config.gpus)
        ),
        addict.Dict(
            name='ports',
            type='RANGES',
            role=role,
            ranges=addict.Dict(range=thaw(task_config.ports)),
        ),
    ]


def make_mesos_command_info(task_config: MesosTaskConfig) -> addict.Dict:
    return addict.Dict(
        value=task_config.cmd,
        uris=[addict.Dict(value=uri, extract=False)
              for uri in task_config.uris],
        environment=make_task_environment_variables(task_config=task_config),
    )


def make_task_environment_variables(task_config: MesosTaskConfig) -> addict.Dict:
    env = dict(task_config.environment.items())
    env['MESOS_TASK_ID'] = task_config.task_id  # type: ignore
    return addict.Dict(variables=[addict.Dict(name=k, value=v) for k, v in env.items()])


def make_mesos_task_info(
    task_config: MesosTaskConfig,
    agent_id: str,
    role: str,
) -> addict.Dict:

    container_info = make_mesos_container_info(task_config)
    resources = make_mesos_resources(task_config, role)
    command_info = make_mesos_command_info(task_config)

    return addict.Dict(
        task_id=addict.Dict(value=task_config.task_id),
        agent_id=addict.Dict(value=agent_id),
        name=f'executor-{task_config.task_id}',
        resources=resources,
        command=command_info,
        container=container_info
    )


MESOS_STATUS_MAP = {
    'TASK_STARTING':
    addict.Dict(platform_type='starting', terminal=False),
    'TASK_RUNNING':
    addict.Dict(platform_type='running', terminal=False),
    'TASK_FINISHED':
    addict.Dict(platform_type='finished', terminal=True, success=True),
    'TASK_FAILED':
    addict.Dict(platform_type='failed', terminal=True, success=False),
    'TASK_KILLED':
    addict.Dict(platform_type='killed', terminal=True, success=False),
    'TASK_LOST':
    addict.Dict(platform_type='lost', terminal=True, success=False),
    'TASK_STAGING':
    addict.Dict(platform_type='staging', terminal=False),
    'TASK_ERROR':
    addict.Dict(platform_type='error', terminal=True, success=False),
    'TASK_KILLING':
    addict.Dict(platform_type='killing', terminal=False),
    'TASK_DROPPED':
    addict.Dict(platform_type='dropped', terminal=True, success=False),
    'TASK_UNREACHABLE':
    addict.Dict(platform_type='unreachable', terminal=False),
    'TASK_GONE':
    addict.Dict(platform_type='gone', terminal=True, success=False),
    'TASK_GONE_BY_OPERATOR':
    addict.Dict(platform_type='gone_by_operator',
                terminal=True, success=False),
    'TASK_UNKNOWN':
    addict.Dict(platform_type='unknown', terminal=False),
    'TASK_STUCK':
    addict.Dict(platform_type='unknown', terminal=False)
}


def mesos_update_to_event(mesos_status: addict.Dict, task_config: MesosTaskConfig) -> Event:
    kwargs = dict(
        raw=mesos_status,
        task_id=task_config.task_id,
        task_config=task_config,
        timestamp=time.time(),
    )
    kwargs.update(MESOS_STATUS_MAP[mesos_status.state])
    return task_event(**kwargs)



import collections
import logging
import time
from queue import Queue
from threading import Lock
from threading import Thread

from task_processing.interfaces.task_executor import TaskExecutor

log = logging.getLogger(__name__)

TaskEntry = collections.namedtuple('TaskEntry', ['task_id', 'deadline'])


class TimeoutExecutor(TaskExecutor):
    def __init__(self, downstream_executor):
        self.downstream_executor = downstream_executor

        self.tasks_lock = Lock()
        # Tasks that are pending termination
        self.killed_tasks = []
        # Tasks that are currently running
        self.running_tasks = []

        self.src_queue = downstream_executor.get_event_queue()
        self.dest_queue = Queue()
        self.stopping = False

        self.timeout_thread = Thread(target=self.timeout_loop)
        self.timeout_thread.daemon = True
        self.timeout_thread.start()

    def timeout_loop(self):
        while True:
            # process downstream events
            while not self.src_queue.empty():
                e = self.src_queue.get()
                self.dest_queue.put(e)

                if not e.kind == 'task':
                    continue
                elif not e.terminal:
                    with self.tasks_lock:
                        if e.task_id not in [entry.task_id for entry in self.running_tasks]:
                            # No record of e's task_id in self.running_tasks,
                            # so we need to add it back in. We lack access to
                            # the original time the task was started, so to set
                            # a deadline, we use e's timestamp as a baseline.
                            new_entry = TaskEntry(
                                task_id=e.task_id,
                                deadline=e.task_config.timeout + e.timestamp,
                            )
                            self._insert_new_running_task_entry(new_entry)
                else:
                    # Update running and killed tasks
                    with self.tasks_lock:
                        for idx, entry in enumerate(self.running_tasks):
                            if e.task_id == entry.task_id:
                                self.running_tasks.pop(idx)
                                break
                        if e.task_id in self.killed_tasks:
                            self.killed_tasks.remove(e.task_id)

            # Check timeouts
            current_time = time.time()
            with self.tasks_lock:
                delete_idx = None
                for idx, entry in enumerate(self.running_tasks):
                    if entry.deadline < current_time:
                        log.info(
                            'Killing task {}: timed out'.format(entry.task_id))
                        self.downstream_executor.kill(entry.task_id)
                        self.killed_tasks.append(entry.task_id)
                        delete_idx = idx
                    # Skip the rest of tasks in the list because they are
                    # appended to the list later.
                    else:
                        break
                if delete_idx is not None:
                    self.running_tasks = self.running_tasks[delete_idx + 1:]

            if self.stopping:
                return

            # Since src_queue has to be polled continuously, sleep(1) is used.
            # Otherwise, a notify() from run() plus wait(delta between now and
            # the earliest deadline) is more efficient.
            time.sleep(1)

    def run(self, task_config):
        # Tasks are dynamically added and removed from running_tasks and
        # and killed_tasks. It's preferable for the client or execution
        # framework to check for duplicated tasks. The duplicate task check does
        # NOT happen here.
        new_entry = TaskEntry(
            task_id=task_config.task_id,
            deadline=task_config.timeout + time.time()
        )
        with self.tasks_lock:
            self._insert_new_running_task_entry(new_entry)

        self.downstream_executor.run(task_config)

    def reconcile(self, task_config):
        self.downstream_executor.reconcile(task_config)

    def kill(self, task_id):
        with self.tasks_lock:
            for idx, entry in enumerate(self.running_tasks):
                if task_id == entry.task_id:
                    log.info('Killing task {}: requested'.format(task_id))
                    result = self.downstream_executor.kill(task_id)
                    if result is not False:
                        self.running_tasks.pop(idx)
                        self.killed_tasks.append(task_id)
                    return result

    def stop(self):
        self.downstream_executor.stop()
        self.stopping = True
        self.timeout_thread.join()

    def get_event_queue(self):
        return self.dest_queue

    def _insert_new_running_task_entry(self, new_entry):
        # Insertion sort for task entries in self.running_tasks
        for idx, entry in enumerate(self.running_tasks):
            if new_entry.deadline <= entry.deadline:
                self.running_tasks.insert(idx, new_entry)
                return
        self.running_tasks.append(new_entry)



TASK_LAUNCHED_COUNT = 'taskproc.mesos.task_launched_count'
TASK_FAILED_TO_LAUNCH_COUNT = 'taskproc.mesos.tasks_failed_to_launch_count'
TASK_LAUNCH_FAILED_COUNT = 'taskproc.mesos.task_launch_failed_count'
TASK_FINISHED_COUNT = 'taskproc.mesos.task_finished_count'
TASK_FAILED_COUNT = 'taskproc.mesos.task_failure_count'
TASK_KILLED_COUNT = 'taskproc.mesos.task_killed_count'
TASK_LOST_COUNT = 'taskproc.mesos.task_lost_count'
TASK_LOST_DUE_TO_INVALID_OFFER_COUNT = 'taskproc.mesos.task_lost_due_to_invalid_offer_count'
TASK_ERROR_COUNT = 'taskproc.mesos.task_error_count'
TASK_OFFER_TIMEOUT = 'taskproc.mesos.task_offer_timeout'

TASK_ENQUEUED_COUNT = 'taskproc.mesos.task_enqueued_count'
TASK_QUEUED_TIME_TIMER = 'taskproc.mesos.task_queued_time'
TASK_INSUFFICIENT_OFFER_COUNT = 'taskproc.mesos.task_insufficient_offer_count'
TASK_STUCK_COUNT = 'taskproc.mesos.task_stuck_count'

OFFER_DELAY_TIMER = 'taskproc.mesos.offer_delay'
BLACKLISTED_AGENTS_COUNT = 'taskproc.mesos.blacklisted_agents_count'

BGCHECK_TIME_TIMER = 'taskproc.mesos.bgcheck_time'



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import find_packages
from setuptools import setup

import task_processing


setup(
    name='task_processing',
    version=task_processing.__version__,
    provides=['task_processing'],
    author='Task Processing',
    author_email='team-taskproc@yelp.com',
    description='Framework for task processing executors and configuration',
    packages=find_packages(exclude=('tests*', "examples*")),
    package_data={
        "task_processing": ["py.typed"]
    },
    install_requires=[
        # used for immutable data structures
        'pyrsistent',
        # until we can drop py36, used for things like TypedDicts
        'typing-extensions',
    ],
    extras_require={
        # We can add the Mesos specific dependencies here
        'mesos_executor': ['addict', 'pymesos>=0.2.14', 'requests'],
        'metrics': ['yelp-meteorite'],
        'persistence': ['boto3'],
        'k8s': ['kubernetes']
    }
)



from .stateful_executor import StatefulTaskExecutor


TASK_PROCESSING_PLUGIN = 'stateful_plugin'


def register_plugin(registry):
    return registry.register_task_executor('stateful', StatefulTaskExecutor)



from .logging_executor import MesosLoggingExecutor
from .mesos_pod_executor import MesosPodExecutor
from .mesos_task_executor import MesosTaskExecutor
from .retrying_executor import RetryingExecutor
from .timeout_executor import TimeoutExecutor


TASK_PROCESSING_PLUGIN = 'mesos_plugin'


def register_plugin(registry):
    return registry \
        .register_task_executor('logging', MesosLoggingExecutor) \
        .register_deprecated_task_executor('mesos', MesosTaskExecutor) \
        .register_task_executor('mesos_task', MesosTaskExecutor) \
        .register_task_executor('mesos_pod', MesosPodExecutor) \
        .register_task_executor('retrying', RetryingExecutor) \
        .register_task_executor('timeout', TimeoutExecutor)



import json

from pyrsistent import thaw
from pyrsistent import v

from task_processing.interfaces.event import Event
from task_processing.interfaces.event import json_deserializer
from task_processing.interfaces.event import json_serializer
from task_processing.interfaces.persistence import Persister


class FilePersistence(Persister):
    def __init__(self, output_file):
        self.output_file = output_file

    def read(self, task_id):
        acc = v()
        with open(self.output_file, 'r') as f:
            for line in f:
                parsed = json.loads(line, object_hook=json_deserializer)
                if parsed['task_id'] == task_id:
                    acc = acc.append(Event.create(parsed))
        return acc

    def write(self, event):
        with open(self.output_file, 'a+') as f:
            f.write("{}\n".format(json.dumps(
                thaw(event), default=json_serializer)))
